#include "stdafx.h"
#include <vd2/system/math.h>
#include "AudioSourceDShow.h"

DSAudioSource::DSAudioSource(const VDXInputDriverContext *context, IDSEngine *engine, const wchar_t *fn)
	: mpEngine(engine)
	, mpContext(context)
{
	memset(&mStreamInfo, 0, sizeof mStreamInfo);
	memset(&mAudioInfo, 0, sizeof mAudioInfo);

	try {
		DSEngineCommand cmd(&mCommandSignal);
		cmd.SetMethodCallback<DSAudioSource, &DSAudioSource::InitLocal>(this);
		cmd.SetExtraData((void *)fn);
		mpEngine->SendCommand(&cmd);
	} catch(const MyError&) {
		ShutdownLocal(NULL);
		throw;
	}
}

DSAudioSource::~DSAudioSource() {
}

// IVDXUnknown
int VDXAPIENTRY DSAudioSource::AddRef() {
	return vdxunknown<IVDXAudioSource>::AddRef();
}

int VDXAPIENTRY DSAudioSource::Release() {
	return vdxunknown<IVDXAudioSource>::Release();
}

void * VDXAPIENTRY DSAudioSource::AsInterface(uint32 iid) {
	if (iid == IVDXStreamSource::kIID)
		return static_cast<IVDXStreamSource *>(this);

	return vdxunknown<IVDXAudioSource>::AsInterface(iid);
}

// IVDXStreamSource
void VDXAPIENTRY DSAudioSource::GetStreamSourceInfo(VDXStreamSourceInfo& info) {
	info = mStreamInfo;
}

bool VDXAPIENTRY DSAudioSource::Read(sint64 lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {
	uint32 samples = lCount;
	uint32 bytes = lCount * mFormat->nBlockAlign;

	if (!lpBuffer) {
		*lBytesRead = bytes;
		*lSamplesRead = samples;
		return true;
	}

	if (bytes > cbBuffer) {
		uint32 bufferInSamples = cbBuffer / mFormat->nBlockAlign;

		if (!bufferInSamples) {
			*lBytesRead = mFormat->nBlockAlign;
			*lSamplesRead = 1;
			return false;
		}

		samples = bufferInSamples;
		bytes = samples * mFormat->nBlockAlign;
	}

	try {
		mReadPosition		= lStart;
		mReadSampleCount	= samples;
		mReadBytesLeft		= bytes;
		mReadBytesActual	= 0;
		mpReadBuffer		= lpBuffer;

		DSEngineCommand cmd(&mCommandSignal);
		cmd.SetMethodCallback<DSAudioSource, &DSAudioSource::DecodeLocal>(this);
		cmd.SetExtraData(NULL);
		mpEngine->SendCommand(&cmd);
	} catch(const MyError& e) {
		mpContext->mpCallbacks->SetError("%s", e.gets());
		return false;
	}

	*lSamplesRead	= mReadBytesActual / mFormat->nBlockAlign;
	*lBytesRead		= mReadBytesActual - mReadBytesActual % mFormat->nBlockAlign;
	return true;
}

const void *VDXAPIENTRY DSAudioSource::GetDirectFormat() {
	return mFormat.data();
}

int VDXAPIENTRY DSAudioSource::GetDirectFormatLen() {
	return mFormat.size();
}

IVDXStreamSource::ErrorMode VDXAPIENTRY DSAudioSource::GetDecodeErrorMode() {
	return IVDXStreamSource::kErrorModeReportAll;
}

void VDXAPIENTRY DSAudioSource::SetDecodeErrorMode(ErrorMode mode) {
}

bool VDXAPIENTRY DSAudioSource::IsDecodeErrorModeSupported(ErrorMode mode) {
	return mode == IVDXStreamSource::kErrorModeReportAll;
}

bool VDXAPIENTRY DSAudioSource::IsVBR() {
	return false;
}

sint64 VDXAPIENTRY DSAudioSource::TimeToPositionVBR(sint64 us) {
	return -1;
}

sint64 VDXAPIENTRY DSAudioSource::PositionToTimeVBR(sint64 samples) {
	return -1;
}

// IVDXAudioSource
void VDXAPIENTRY DSAudioSource::GetAudioSourceInfo(VDXAudioSourceInfo& info) {
	info = mAudioInfo;
}

void DSAudioSource::InitLocal(void *pvFn) {
	const wchar_t *fn = (const wchar_t *)pvFn;

	HRESULT hr = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER, IID_IGraphBuilder, (void **)~mpGraph);
	if (FAILED(hr))
		throw MyError("Impossible de crer le graph du filtre DirectShow: hr=%08x", hr);

	hr = mpGraph->QueryInterface(IID_IMediaControl, (void **)~mpGraphMediaControl);
	if (FAILED(hr))
		throw MyError("Impossible d'obtenir l'interface IMediaControl DirectShow: hr=%08x", hr);

	hr = mpGraph->QueryInterface(IID_IMediaEvent, (void **)~mpGraphMediaEvent);
	if (FAILED(hr))
		throw MyError("Impossible d'obtenir l'interface IMediaEvent DirectShow: hr=%08x", hr);

	hr = mpGraph->QueryInterface(IID_IMediaSeeking, (void **)~mpGraphMediaSeeking);
	if (FAILED(hr))
		throw MyError("Impossible d'obtenir l'interface IMediaSeeking DirectShow: hr=%08x", hr);

	hr = mpGraph->QueryInterface(IID_IMediaFilter, (void **)~mpGraphMediaFilter);
	if (FAILED(hr))
		throw MyError("Impossible d'obtenir l'interface IMediaFilter DirectShow: hr=%08x", hr);

	// kill the graph clock
	hr = mpGraphMediaFilter->SetSyncSource(NULL);
	if (FAILED(hr))
		throw MyError("Impossible de dsactiver l'horloge du filtre DirectShow: hr=%08x", hr);

	mpSampleGrabber = new_nothrow DSSampleGrabber;
	if (!mpSampleGrabber)
		throw MyMemoryError();

	mpSampleGrabber->SetCallback(this);
	mpSampleGrabber->SetMediaType(&MEDIATYPE_Audio);

	hr = mpGraph->AddFilter(mpSampleGrabber, L"audio grabber");
	if (FAILED(hr))
		throw MyError("Impossible d'ajouter un grabber d'chantillons au graph: hr=%08x", hr);

	vdrefptr<ICaptureGraphBuilder2> pCB2;
	hr = CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, CLSCTX_INPROC_SERVER, IID_ICaptureGraphBuilder2, (void **)~pCB2);
	if (FAILED(hr))
		throw MyError("Impossible de crer le gestionnaire de graph de capture: hr=%08x", hr);

	hr = pCB2->SetFiltergraph(mpGraph);
	if (FAILED(hr))
		throw MyError("Impossible de dfinir le graph des filtres sur le gestionnaire de graph de capture: hr=%08x", hr);

	vdrefptr<IBaseFilter> pSourceFilter;
	hr = mpGraph->AddSourceFilter(fn, L"source", ~pSourceFilter);
	if (FAILED(hr))
		throw MyError("Impossible de crer le filtre source: hr=%08x", hr);

	hr = CoCreateInstance(CLSID_NullRenderer, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (void **)~mpRenderer);
	if (FAILED(hr))
		throw MyError("Impossible de crer un rendu null: hr=%08x", hr);

	hr = mpGraph->AddFilter(mpRenderer, L"null renderer");
	if (FAILED(hr))
		throw MyError("Impossible d'ajouter le rendu null au graph: hr=%08x", hr);

	{
		vdrefptr<IPin> srcpin;
		hr = pCB2->FindPin(pSourceFilter, PINDIR_OUTPUT, NULL, &MEDIATYPE_Audio, TRUE, 0, ~srcpin);
		if (FAILED(hr)) {
			hr = pCB2->RenderStream(NULL, NULL, pSourceFilter, mpSampleGrabber, mpRenderer);
			if (FAILED(hr))
				throw MyError("Impossible de faire un rendu du flux audio: hr=%08x", hr);
		} else {
			hr = mpGraph->Connect(srcpin, mpSampleGrabber->GetInputPin());
			if (FAILED(hr))
				throw MyError("Impossible de faire un rendu du flux audio: hr=%08x", hr);

			vdrefptr<IPin> rdrpin;
			hr = pCB2->FindPin(mpRenderer, PINDIR_INPUT, NULL, NULL, TRUE, 0, ~rdrpin);
			if (FAILED(hr))
				throw MyError("Impossible de faire un rendu du flux audio: hr=%08x", hr);

			hr = mpGraph->Connect(mpSampleGrabber->GetOutputPin(), rdrpin);
			if (FAILED(hr))
				throw MyError("Impossible de faire un rendu du flux audio: hr=%08x", hr);
		}
	}

	hr = mpSampleGrabber->GetInputPin()->ConnectedTo(~mpSourcePin);
	if (FAILED(hr))
		throw MyError("Impossible de trouver une pin de sortie source/dcodeur: hr=%08x", hr);

	hr = mpSampleGrabber->GetOutputPin()->ConnectedTo(~mpRendererPin);
	if (FAILED(hr))
		throw MyError("Impossible de trouver une pin d'entre pour le rendu: hr=%08x", hr);

	// must get duration before we disconnect the source pin
	LONGLONG dur;
	hr = mpGraphMediaSeeking->GetDuration(&dur);
	if (FAILED(hr))
		throw MyError("Impossible d'obtenir la dure du flux: hr=%08x", hr);

	hr = mpSourcePin->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(SUCCEEDED(hr));

	hr = mpSampleGrabber->GetInputPin()->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(SUCCEEDED(hr));

	hr = mpSampleGrabber->GetOutputPin()->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(SUCCEEDED(hr));

	vdrefptr<IEnumMediaTypes> pMediaTypes;
	hr = mpSourcePin->EnumMediaTypes(~pMediaTypes);
	if (FAILED(hr))
		throw MyError("Impossible d'numrer les types de mdia: hr=%08x", hr);

	AM_MEDIA_TYPE *amt;
	ULONG fetched;
	int rate;
	bool found = false;

	while(S_OK == pMediaTypes->Next(1, &amt, &fetched)) {
		if (amt->majortype == MEDIATYPE_Audio) {
			if (amt->formattype == FORMAT_WaveFormatEx) {
				const WAVEFORMATEX& wfex = *(const WAVEFORMATEX *)amt->pbFormat;

				rate = wfex.nSamplesPerSec;

				mFormat.assign(&wfex, amt->cbFormat);

				mStreamInfo.mSampleRate.mNumerator		= wfex.nSamplesPerSec;
				mStreamInfo.mSampleRate.mDenominator	= 1;
				mStreamInfo.mPixelAspectRatio.mNumerator	= 0;
				mStreamInfo.mPixelAspectRatio.mDenominator	= 0;

				found = true;
			}
		}

		if (amt->pbFormat)
			CoTaskMemFree(amt->pbFormat);
		CoTaskMemFree(amt);

		if (found)
			break;
	}

	if (!found)
		throw MyError("Impossible de trouver un format de dcompression compatible.");

	mStreamInfo.mSampleCount = VDRoundToInt64((double)rate * (double)dur / 10000000.0);
}

void DSAudioSource::ShutdownLocal(void *) {
	mpRendererPin = NULL;
	mpRenderer = NULL;
	mpSampleGrabber = NULL;
	mpSourcePin = NULL;
	mpGraphMediaSeeking = NULL;
	mpGraphMediaEvent = NULL;
	mpGraphMediaControl = NULL;
	mpGraphMediaFilter = NULL;
	mpGraph = NULL;
}

void DSAudioSource::DecodeLocal(void *) {
	// change time on sample grabber and see if it can grab the frame from its queue
	LONGLONG timeStartFlush = VDRoundToInt64((double)mReadPosition * (double)mStreamInfo.mSampleRate.mDenominator / (double)mStreamInfo.mSampleRate.mNumerator * 10000000.0);
	LONGLONG timeStart = VDRoundToInt64(((double)mReadPosition + 0.5) * (double)mStreamInfo.mSampleRate.mDenominator / (double)mStreamInfo.mSampleRate.mNumerator * 10000000.0);
	LONGLONG timeEnd = VDRoundToInt64(((double)(mReadPosition + mReadSampleCount) + 0.5) * (double)mStreamInfo.mSampleRate.mDenominator / (double)mStreamInfo.mSampleRate.mNumerator * 10000000.0);

	mReadTimeStart = timeStartFlush;

	if (mpSampleGrabber->CheckForSample(timeStart, timeEnd)) {
		// sweet
		return;
	}

	// issue a seek
	HRESULT hr;
	OAFilterState fs;

	REFERENCE_TIME lastSeekPos;
	if (!mpSampleGrabber->GetEstimatedStreamPosition(&lastSeekPos) || timeStart < lastSeekPos || timeStart > lastSeekPos + 5000000) {
		LONGLONG startPos = timeStartFlush;

		// nope... issue a pause
		hr = mpGraphMediaControl->Pause();
		if (FAILED(hr))
			throw MyError("Impossible de mettre en pause le graph DirectShow: code d'erreur %08x", hr);

		// wait for pending state changes to complete
	#if 0
		hr = mpGraphMediaControl->GetState(INFINITE, &fs);
		if (FAILED(hr))
			throw MyError("Impossible de mettre en pause le graph DirectShow: code d'erreur %08x", hr);
	#endif

		// lock down the sample grabber
		mpSampleGrabber->ResetOneShot(timeStart, timeEnd);

		hr = mpGraphMediaSeeking->SetPositions(&startPos, AM_SEEKING_AbsolutePositioning, NULL, AM_SEEKING_NoPositioning);
		if (FAILED(hr))
			throw MyError("Impossible de rechercher dans le graph DirectShow: code d'erreur %08x", hr);
	} else {
		if (mpSampleGrabber->ResumeOneShot(timeStart, timeEnd))
			return;
	}

	// run the graph
	hr = mpGraphMediaControl->GetState(INFINITE, &fs);
	if (FAILED(hr))
		throw MyError("Impossible de lancer le graph DirectShow: code d'erreur %08x", hr);

	if (fs != State_Running) {
		hr = mpGraphMediaControl->Run();
		if (FAILED(hr))
			throw MyError("Impossible de lancer le graph DirectShow: code d'erreur %08x", hr);

		// wait for state transition to complete
		hr = mpGraphMediaControl->GetState(INFINITE, &fs);
		if (FAILED(hr))
			throw MyError("Impossible de lancer le graph DirectShow: code d'erreur %08x", hr);
	}

	// wait for sample
	mpSampleGrabber->WaitForSample();

	mpReadBuffer = NULL;
	mReadBytesLeft = 0;
}

bool DSAudioSource::Process(IMediaSample *pSample, const REFERENCE_TIME& timeBase) {
	REFERENCE_TIME t1, t2;
	HRESULT hr = pSample->GetTime(&t1, &t2);
	VDDEBUG("Receiving audio: [%d, %d]\n", (int)(t1 + timeBase), (int)(t2 + timeBase));

	REFERENCE_TIME trelstart = mReadTimeStart - timeBase;
	int offsetInBytes = 0;
	if (SUCCEEDED(hr) && t1 < trelstart) {
		REFERENCE_TIME tdelta = trelstart - t1;
		int samples = VDRoundToInt((double)tdelta / (double)mStreamInfo.mSampleRate.mDenominator * (double)mStreamInfo.mSampleRate.mNumerator / 10000000.0);
		int bytes = samples * mFormat->nBlockAlign;

		offsetInBytes = bytes;
	}

	BYTE *p;
	hr = pSample->GetPointer(&p);

	if (FAILED(hr))
		return false;

	long size = pSample->GetActualDataLength();
	VDASSERT(size >= 0);
	if (size <= 0)
		return true;

	if (offsetInBytes >= size) {
		VDASSERT(offsetInBytes == size);
		return true;
	}

	uint32 usize = size - offsetInBytes;
	if (usize > mReadBytesLeft)
		usize = mReadBytesLeft;

	memcpy(mpReadBuffer, p + offsetInBytes, usize);
	mpReadBuffer = (char *)mpReadBuffer + usize;
	mReadBytesLeft -= usize;
	mReadBytesActual += usize;

	return mReadBytesLeft > 0;
}
