#include "stdafx.h"
#include <vd2/system/thread.h>
#include <vd2/system/w32assist.h>

#include "Engine.h"

#include <vd2/system/event.h>

class DSEngine : public VDThread, public vdrefcounted<IDSEngine> {
public:
	typedef DSEngineCommand Command;

	DSEngine();
	~DSEngine();

	bool Init();
	void Shutdown();

	void PostCommand(Command *cmd);
	void WaitCommand(Command *cmd);
	void SendCommand(Command *cmd);

	void Finish(VDSignal *signal);

protected:
	bool DispatchCommands();

	void ThreadRun();

	bool InitLocal();
	void ShutdownLocal();

	VDAtomicInt			mLocallyInited;
	VDAtomicInt			mCommandsQueued;
	VDCriticalSection	mLock;

	vdlist<Command>		mCommandQueue;
};

bool VDCreateDSEngine(IDSEngine **pp) {
	*pp = new DSEngine;
	(*pp)->AddRef();
	return true;
}

DSEngine::DSEngine()
	: mLocallyInited(0)
	, mCommandsQueued(0)
{
}

DSEngine::~DSEngine() {
}

bool DSEngine::Init() {
	if (!ThreadStart()) {
		Shutdown();
		return false;
	}

	// wait for thread to init
	VDSignal tempSignal;
	Finish(&tempSignal);

	// check if init was successful
	if (!mLocallyInited) {
		Shutdown();
		return false;
	}

	return true;
}

void DSEngine::Shutdown() {
	if (isThreadActive()) {
		PostThreadMessage(getThreadID(), WM_QUIT, 0, 0);
		ThreadWait();
	}
}

void DSEngine::PostCommand(Command *cmd) {
	VDASSERT(!cmd->mCompleted);

	vdsynchronized(mLock) {
		mCommandQueue.push_back(cmd);
	}

	if (!mCommandsQueued++)
		PostThreadMessage(getThreadID(), WM_NULL, 0, 0);
}

void DSEngine::WaitCommand(Command *cmd) {
	if (cmd->mpSignal) {
		// normal path
		HANDLE h[2] = {cmd->mpSignal->getHandle(), getThreadHandle()};

		while(!cmd->mCompleted) {
			// Wait on both the signal and the thread, in case we have some catastrophic startup
			// failure.
			DWORD result = WaitForMultipleObjects(2, h, FALSE, INFINITE);

			if (result != WAIT_OBJECT_0)
				break;
		}
	} else {
		// fallback path
		while(!cmd->mCompleted && isThreadActive())
			::Sleep(10);
	}
}

void DSEngine::SendCommand(Command *cmd) {
	PostCommand(cmd);
	WaitCommand(cmd);
	cmd->ThrowError();
}

void DSEngine::Finish(VDSignal *signal) {
	Command fenceCmd(signal);

	PostCommand(&fenceCmd);
	WaitCommand(&fenceCmd);
}

void DSEngine::ThreadRun() {
	MSG msg;

	if (!InitLocal())
		return;

	for(;;) {
		while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT)
				goto xit;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		if (!DispatchCommands())
			WaitMessage();
	}

xit:
	ShutdownLocal();
}

bool DSEngine::DispatchCommands() {
	if (!mCommandsQueued)
		return false;

	--mCommandsQueued;

	Command *cmd;
	vdsynchronized(mLock) {
		VDASSERT(!mCommandQueue.empty());
		cmd = mCommandQueue.front();
		mCommandQueue.pop_front();
	}

	cmd->mError.clear();
	try {
		if (cmd->mpCallback)
			cmd->mpCallback(cmd->mpData, cmd->mpCallbackData);
	} catch(MyError& e) {
		cmd->mError.TransferFrom(e);
	}

	cmd->mCompleted = true;
	if (cmd->mpSignal)
		cmd->mpSignal->signal();

	return true;
}

bool DSEngine::InitLocal() {
	// bring up Direct3D 9
	HRESULT hr = CoInitialize(NULL);
	if (FAILED(hr)) {
		ShutdownLocal();
		return false;
	}

	mLocallyInited = true;
	return true;
}

void DSEngine::ShutdownLocal() {
	if (mLocallyInited) {
		mLocallyInited = false;
		CoUninitialize();
	}
}
